--------------------------------------------------------------------------------
--   ____  ____
--  /   /\/   /
-- /___/  \  /   Vendor: Xilinx
-- \   \   \/    Version: 1.0
--  \   \        Filename: $RCSfile: circ_buffer_dec.vhd,v $           
--  /   /        Date Last Modified: $Date: 2008/08/14 14:01:45 $ 
-- /___/   /\    Date Created: April 2008 
-- \   \  /  \   
--  \___\/\___\
-- 
--
-------------------------------------------------------------------------------- 
-- DISCLAIMER OF LIABILITY
-- 
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you 
-- a license to use this text/file solely for design, simulation, 
-- implementation and creation of design files limited 
-- to Xilinx devices or technologies. Use with non-Xilinx 
-- devices or technologies is expressly prohibited and 
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information 
-- "as is" solely for use in developing programs and 
-- solutions for Xilinx devices. By providing this design, 
-- code, or information as one possible implementation of 
-- this feature, application or standard, Xilinx is making no 
-- representation that this implementation is free from any 
-- claims of infringement. You are responsible for 
-- obtaining any rights you may require for your implementation. 
-- Xilinx expressly disclaims any warranty whatsoever with 
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied 
-- warranties of merchantability or fitness for a particular 
-- purpose. 
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
-- 
--
-- Copyright (c) 2008 Xilinx, Inc. All rights reserved.
--
-- This copyright and support notice must be retained as part 
-- of this text at all times. 
--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
-- Description:
-- BRAM based circular buffer
-- for reading out training blocks for Viterbi decoding
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.global_types_pkg.all;

entity circ_buffer_dec is
  generic
  (
    c_addr_width : integer := 4; -- buffer depth width
    c_data_width : integer := 1 -- buffer width
  );
  port
  (
	  -- data input handling
    data_in     : in std_logic_vector(c_data_width-1 downto 0); -- data written to buffer sequentially
    fd_wr       : in std_logic; -- first data start of buffer block -- only valid when nd and rfd are high
    ld_wr       : in std_logic; -- last data end of buffer block -- only valid when nd and rfd are high
    nd          : in std_logic; -- input data is valid if rfd is high
    rfd         : out std_logic; --  data can be written
    -- data output handling
    cts         : in std_logic;
    data_out    : out std_logic_vector(c_data_width-1 downto 0);
    block_out   : out std_logic; -- marks viterbi block
    -- clocks etc
    clk         : in std_logic;
	  ext2all     : in GLB_InterfaceEXT2ALL  := GLB_InterfaceEXT2ALL_DEFAULT
   );
end circ_buffer_dec;

architecture synth of circ_buffer_dec is

constant RAM_DELAY : positive := 3;
constant BUF_DEPTH : natural := 2**C_ADDR_WIDTH;
constant FIFO_W : natural := 2*C_ADDR_WIDTH;
constant BLOCK_DELAY : positive := RAM_DELAY;

  type FIFO_OFF_R is record
   offset     : std_logic_vector(C_ADDR_WIDTH-1 downto 0);
   data_in    : std_logic_vector(FIFO_W-1 downto 0);
   we         : std_logic;
   data_out   : std_logic_vector(FIFO_W-1 downto 0);
   re         : std_logic;
   re_cts     : std_logic;
   rd_valid   : std_logic;
   full       : std_logic;
   empty      : std_logic;
  end record;  

  type RAM_RD_STATE is (START_RD, WAIT_OFFSETS, GET_OFFSETS,GET_START, COST_DATA, VIT_DATA,TB_DATA);

  type RAM_R is record
    data_in     : std_logic_vector(C_DATA_WIDTH-1 downto 0);
    waddr       : unsigned(C_ADDR_WIDTH-1 downto 0);
    we          : std_logic;

    rd_state    : RAM_RD_STATE;
    raddr       : unsigned(C_ADDR_WIDTH-1 downto 0);
    start_offset: unsigned(C_ADDR_WIDTH-1 downto 0);
    end_offset  : unsigned(C_ADDR_WIDTH-1 downto 0);
    count_data  : unsigned(TB_W-1 downto 0); -- for counting the training data
    length      : unsigned(C_ADDR_WIDTH-1 downto 0);
    diff        : unsigned(C_ADDR_WIDTH-1 downto 0);
    re          : std_logic;

    block_out   : std_logic;
    data_out    : std_logic_vector(C_DATA_WIDTH-1 downto 0);
    block_out_v : std_logic_vector(BLOCK_DELAY-1 downto 0);
  end record;
  
  

signal fifo_d : FIFO_OFF_R;  
signal ram_d : RAM_R;
signal enb : std_logic;
signal fd_wr_int,ld_wr_int,nd_int : std_logic;
signal fd_wr_next,ld_wr_next : std_logic;
signal rfd_int : std_logic;

begin
-----------------------------------------------------------------------------
-- make sure signals are valid
-----------------------------------------------------------------------------
  nd_int <= nd and rfd_int;
  fd_wr_int <= fd_wr and nd_int; 
  ld_wr_int <= ld_wr and nd_int;

-----------------------------------------------------------------------------
--  block input addr offsets to fifo
-- offsets are caught on the fd_wr and saved on the ld_wr
-- offset is stored on the ld_wr therefore only entry in fifo if the block
-- has been written
-- 16 deep and 2*BRAM address width wide -- stores both the start and end address
-- for the block
-- used by the raddr to move throught the blocks
-----------------------------------------------------------------------------
  fifo_d.re_cts <= fifo_d.re and cts;
  
  block_offset: entity work.fifo_wrap 
	generic map
	(
		C_WIDTH   => FIFO_W,
    C_DEPTH_W => FIFO_DEPTH_W
	)
	port map
  (
    clk      => clk,
    rst      => ext2all.reset,  

    din      => fifo_d.data_in,  
    wr_en    => fifo_d.we,

    dout     => fifo_d.data_out, 
    rd_en    => fifo_d.re_cts,
    valid    => fifo_d.rd_valid,

    empty    => fifo_d.empty,
    full     => fifo_d.full
   );

------------------------------------------------------------------------
-- data write process
-- write increments sequentially for each valid nd ( conditional upon rfd)
-- and saves offsets to the fifo, writes only halted when fifo or ram 
-- is full
------------------------------------------------------------------------

 	data_wr: process(clk)
	begin
    if rising_edge(clk) and ext2all.ce='1' then
      -- Synchronous reset
      if ext2all.reset = '1' then
        ram_d.waddr <= (others => '1');
        ram_d.we <= '0';
        ram_d.data_in <= (others => '0');
        fifo_d.offset <= (others => '0');
        fifo_d.data_in <= (others => '0');
        fifo_d.we <= '0';
        rfd_int <= '0';
        fd_wr_next <= '0';
        ld_wr_next <= '0';
      else
        ram_d.data_in <= data_in;
        -- rfd low only when buffer is completely full
        if(ram_d.waddr = ram_d.start_offset ) then
          rfd_int <= '0';
        else
          rfd_int <= '1';
        end if;
        -- write data when nd_int is high, 
        -- note that nd_int is qualified by rfd_int
        if(nd_int = '1') then
          fd_wr_next <= fd_wr_int;
          ld_wr_next <= ld_wr_int;
          ram_d.we <= '1';
          ram_d.waddr <= ram_d.waddr + 1;
--          assert not(ram_d.waddr = 0)
--             report "*************** : Circular buffer wrap" 
--              severity note; 
        else
          fd_wr_next <= '0';
          ld_wr_next <= '0';
          ram_d.we <= '0';
        end if; 
        -- hold start offset for fifo when fd is writen
        if(fd_wr_next = '1') then
          fifo_d.offset <= std_logic_vector(ram_d.waddr);
        end if;    
        -- write block start address and block end address to fifo when block is complete
        if(ld_wr_next = '1') then
          fifo_d.data_in <= std_logic_vector(ram_d.waddr) & fifo_d.offset ;
          fifo_d.we <= '1';
        else
          fifo_d.we <= '0';
        end if;    
      end if;
    end if;  
  end process data_wr;    

------------------------------------------------------------------------
-- data read process
-- 
-- by current block offset for reading within the circular buffer
------------------------------------------------------------------------

 	data_rd: process(clk)
	begin
    if rising_edge(clk) and ext2all.ce='1' then
      -- Synchronous reset
      if ext2all.reset = '1' then
        -- ram control
        ram_d.raddr <= (others => '0');
        ram_d.start_offset(C_ADDR_WIDTH-1 downto 1) <= (others => '1') ;
        ram_d.start_offset(0) <= '0';
        ram_d.end_offset <= (others => '0');
        ram_d.rd_state <= START_RD;
        ram_d.count_data <= (others => '0');
        ram_d.length <= (others => '0');
        ram_d.diff <= (others => '0');
        ram_d.re <= '0';
        ram_d.block_out_v <= (others => '0');
        ram_d.block_out <= '0';
        
        -- fifo read
        fifo_d.re <= '0';        
      else
        if(cts = '1') then
          ram_d.block_out_v(BLOCK_DELAY-1 downto 1) <= ram_d.block_out_v(BLOCK_DELAY-2 downto 0);
          ram_d.block_out_v(0) <= ram_d.block_out;
          case ram_d.rd_state is
            -- at start wait until a block is available
            -- indicated by the fifo not empty flag
            -- then move to getting the offsets
            when START_RD =>
              -- can only read when the full buffer is there
              if(fifo_d.empty = '0') then
                fifo_d.re <= '1';
                ram_d.rd_state  <= WAIT_OFFSETS;
              else
                fifo_d.re <= '0';
                ram_d.rd_state  <= START_RD;
              end if;
              ram_d.re <= '0';
              ram_d.count_data <= (others => '0');
              ram_d.length <= (others => '0');
              ram_d.diff <= (others => '0');
              ram_d.start_offset <= ram_d.start_offset;
              ram_d.end_offset <= ram_d.end_offset;
              ram_d.raddr <= (others => '0');
              ram_d.block_out <= '0';
              
            -- fifo has been read pull the data into the start and end offsets
            -- need a delay for the data to appear on output of fifo hence wait and get
            -- and move to reading out the cost training buffer
            when WAIT_OFFSETS =>
              ram_d.rd_state  <= GET_OFFSETS;
              ram_d.start_offset <= ram_d.start_offset;
              ram_d.end_offset <= ram_d.end_offset;
              fifo_d.re <= '0';
              ram_d.count_data <= (others => '0');
              ram_d.length <= (others => '0');
              ram_d.diff <= (others => '0');
              ram_d.raddr <= (others => '0');
              ram_d.re <= '0';
              ram_d.block_out <= '0';
            when GET_OFFSETS =>
              ram_d.rd_state  <= GET_START;
              ram_d.start_offset <= unsigned(fifo_d.data_out(C_ADDR_WIDTH-1 downto 0));
              ram_d.end_offset <= unsigned(fifo_d.data_out(2*C_ADDR_WIDTH-1 downto C_ADDR_WIDTH));
              ram_d.count_data <= (others => '0');
              fifo_d.re <= '0';
              ram_d.raddr <= (others => '0');
              ram_d.length <= to_unsigned(TRACEBACK_LENGTH-1,C_ADDR_WIDTH);
              ram_d.diff <=  unsigned(fifo_d.data_out(2*C_ADDR_WIDTH-1 downto C_ADDR_WIDTH))-unsigned(fifo_d.data_out(C_ADDR_WIDTH-1 downto 0))+1;
              ram_d.re <= '0';
              ram_d.block_out <= '0';

            -- figure out the start address  
            -- and place in raddr
            when GET_START =>
              if(ram_d.length > ram_d.diff) then
                ram_d.length <= ram_d.length - ram_d.diff;
                ram_d.re <= '0';
                ram_d.rd_state  <= GET_START;
              else
                ram_d.raddr <= ram_d.end_offset - ram_d.length;
                ram_d.re <= '1';
                ram_d.rd_state  <= COST_DATA;
              end if;               
              fifo_d.re <= '0';
              ram_d.start_offset <= ram_d.start_offset;
              ram_d.end_offset <= ram_d.end_offset;
              ram_d.count_data <= (others => '0');
              ram_d.block_out <= '0';
              
            -- read out the TRACEBACK length of data for viterbi prior to main block
            when COST_DATA =>
              ram_d.count_data <= ram_d.count_data +1;
              fifo_d.re <= '0';
              ram_d.start_offset <= ram_d.start_offset;
              ram_d.end_offset <= ram_d.end_offset;
              if(ram_d.count_data >= to_unsigned(TRACEBACK_LENGTH-1,TB_W)) then
                ram_d.raddr <= ram_d.start_offset;
                ram_d.count_data <= (others => '0');
                ram_d.rd_state <= VIT_DATA;
                ram_d.block_out <= '1';
              else  
                if(ram_d.raddr = ram_d.end_offset) then
                  ram_d.raddr <= ram_d.start_offset;
                else  
                  ram_d.raddr <= ram_d.raddr + 1;
                end if;
                ram_d.block_out <= '0';
                ram_d.rd_state <= COST_DATA;
              end if;
              ram_d.re <= '1';
            
            -- read out the data for the viterbi sequentially
            -- stop when we reach the end of the block address  
            when VIT_DATA =>
              fifo_d.re <= '0';
              ram_d.start_offset <= ram_d.start_offset;
              ram_d.end_offset <= ram_d.end_offset;
              ram_d.count_data <= (others => '0');
              ram_d.raddr <= ram_d.raddr + 1;
              if(ram_d.raddr = ram_d.end_offset) then
                ram_d.block_out <= '0';
                ram_d.rd_state <= TB_DATA;
                ram_d.raddr <= ram_d.start_offset;
                ram_d.re <= '0';
              else
                ram_d.block_out <= '1';
                ram_d.rd_state <= VIT_DATA;
                ram_d.re <= '1';
              end if;    
              
            -- read out the training data for the viterbi sequentially
            -- start at the start of the block and count
            when TB_DATA =>
              ram_d.count_data <= ram_d.count_data +1;
              if(ram_d.count_data >= to_unsigned(TRACEBACK_LENGTH-1,TB_W)) then
                ram_d.raddr <= ram_d.start_offset;
                ram_d.count_data <= (others => '0');
                ram_d.rd_state <= START_RD;
              else  
                if(ram_d.raddr = ram_d.end_offset) then
                  ram_d.raddr <= ram_d.start_offset;
                else  
                  ram_d.raddr <= ram_d.raddr + 1;
                end if;
                ram_d.rd_state <= TB_DATA;
              end if;
              
          
            when others =>
              fifo_d.re <= '0';
              ram_d.start_offset <= (others => '0');
              ram_d.end_offset <= (others => '0');
              ram_d.raddr <= (others => '0');
              ram_d.length <= (others => '0');
              ram_d.count_data <= (others => '0');
              ram_d.diff <= (others => '0');
              ram_d.rd_state <= START_RD;
              ram_d.re <= '0';
              
          end case;
        end if;          
      end if;
    end if;  
  end process data_rd;    

------------------------------------------------------------------------
-- instantiate BRAM for circular buffer
-- depth is > 2*6144
-- width is 1 to hold the systematic data prior to encoding
-- this could probably be in dist mem if we have an upper limit
-- on the control block length for the CCH channel
------------------------------------------------------------------------
    enb <= ext2all.ce and cts;

    dp_ram: entity work.glb_dpm 
    generic map(
      width            => C_DATA_WIDTH,
      depth            => C_ADDR_WIDTH,
      delay            => RAM_DELAY
      )  
    port map (
      addra  => std_logic_vector(ram_d.waddr),
      addrb  => std_logic_vector(ram_d.raddr),
      clk    => clk,
      wea    => ram_d.we,
      ena    => ext2all.ce,
      enb    => enb,
      dina   => ram_d.data_in,
      doutb  => ram_d.data_out
     );



------------------------------------------------------------------------
-- Output handshaking
-- 
------------------------------------------------------------------------
  RFD <= rfd_int and not(fifo_d.full);
  DATA_OUT <= ram_d.data_out;
  BLOCK_OUT <= ram_d.block_out_v(BLOCK_DELAY-3);


end synth;